class AnilistObject:
    """An object representing data sent by the Anilist API
    
    .. warning:: 
    
        This should never be created manually

    .. note::
        
        The attributes for any specific :class:`~AnilistObject` are the same as for the :class:`~.scheme.Scheme` provided in the request

    Example:

    .. code-block:: python

        ###! In a raw Client request
        client = Anilist.QueryClient()
        query = \"\"\"
            Media (type: $type, search: $search) {
                id
                title {
                    english
                }
            }
        \"\"\"
        result = client._request(
            query,
            vars=Vars(type=AnilistMediaType("ANIME"), search="One Piece")
        )
        # result is an AnilistObject with the following attributes
        id = result.Media.id
        english_title = result.Media.title.english

        ###! In a basic Client request
        client = Anilist.QueryClient()
        schs = Scheme().Media(type="$type", search="$search").sub_schemes(
            Scheme().id,
            Scheme().title.english
        )
        vars = Vars(type=AnilistMediaType("ANIME"), search="One Piece")
        result = client._request(client._create_query(vars, *schs), vars)

        # result is an AnilistObject with the following attributes
        id = result.Media.id
        english_title = result.Media.title.english

        

        ###! Using the Query API
        media_search = Anilist.QueryClient().media_entry()
        media_search.search(
            Scheme().id,
            Scheme().title.english,

            type=AnilistMediaType("ANIME"),
            search="One Piece"
        )
        result = media_search.results_take_one()

        # The Query API automatically returns 
        # only the inner part of the request
        # so you do not need to do 'result.Media'
        # result is an AnilistObject with the following attributes
        id = result.id
        english_title = result.title.english
        # This one is added because the MediaEntryQuery API 
        # has coverImage.extraLarge as a default field to query
        extra_large_cover = result.coverImage.extraLarge
        
    """
    
    def __init__(self, json_object):
        self.__dict = {}
        for k, v in json_object.items():
            if type(v) is dict:
                self.__dict[k] = AnilistObject(v)
            elif type(v) is list:
                self.__dict[k] = [AnilistObject(el) if type(el) is dict else el for el in v]
            else:
                self.__dict[k] = v

    def __getattr__(self, k):
        return self.__dict[k]
    
    def __getitem__(self, k):
        return self.__dict[k]
    
    def __str__(self):
        return str(self.__dict)
    
    def __repr__(self):
        return str(self)


from Anilist.client.client import Client
from Anilist.mutation.mutable import Mutable
from Anilist.scheme import Scheme
from Anilist.vars import Vars


class MediaEntryMutable(Mutable):
    
    def __init__(self, id, client: Client):
        self._id = id
        self._client = client

    def set_score(self, score: float):

        self._mutate(Scheme().id, Scheme().score, head_sch=Scheme().SaveMediaListEntry, id=self._id, score=score)

        return "OK"
    
    def delete(self):
        
        self._mutate(Scheme().deleted, head_sch=Scheme().DeleteMediaListEntry, id=self._id)

        return "OK"

    @classmethod
    def _from_media_id(cls, client: Client, media_id):
        vars = Vars(mediaId=media_id)
        q = client._create_query(vars, Scheme().id, head_sch=Scheme().SaveMediaListEntry(mediaId='$mediaId'))

        resp = client._request(q, vars)

        return cls(resp.SaveMediaListEntry.id, client)



from Anilist.scheme import Scheme
from Anilist.vars import Vars


class Mutable:

    def __init__(self, client):
        self._client = client
    
    def _mutate(self, *schs: Scheme, head_sch: Scheme, **kwargs):
        
        v = Vars(**kwargs)
        q = self._client._create_query(v, *schs, head_sch=head_sch(**v._as_query()))

        return self._client._request(q, v)


from .scheme import Scheme

class MediaScheme(Scheme):
    def __init__(self, **kwargs):
        super().__init__()
        self._layers.append("Media")
        self(**kwargs)

class mediaScheme(Scheme):
    def __init__(self, **kwargs):
        super().__init__()
        self._layers.append("media")
        self(**kwargs)

class mediaListScheme(Scheme):
    def __init__(self, **kwargs):
        super().__init__()
        self._layers.append("mediaList")
        self(**kwargs)

class MediaListScheme(Scheme):
    def __init__(self, **kwargs):
        super().__init__()
        self._layers.append("MediaList")
        self(**kwargs)


from Anilist.client import Client
from Anilist.scheme.scheme import Scheme
from Anilist.vars.vars import Vars


class Query:
    """An interface to make Queries to the Anilist API
    This should usually not be used, instead use one of the subclasses

    .. warning::

        This usually should not be created manually

    :param client: A :class:`~Anilist.client.QueryClient` to make requests
    :type client: :class:`~.client.Client`
        
    :param head_name1: Name of the head :class:`~.scheme.Scheme` when in a Page object
    :type head_nama: :class:`~str`

    :param head_name2: Name of the head :class:`~.scheme.Scheme` when in a Query object
    :type head_namb: :class:`~str`

    :param defa: List of default :class:`~.scheme.Scheme`\s 
    :type defa: list[:class:`~.scheme.Scheme`]

    Example:

    .. code-block:: python

        
        client = Anilist.QueryClient()
        schs = Scheme().Media(type="$type", search="$search").sub_schemes(
            Scheme().id,
            Scheme().title.english
        )
        vars = Vars(type=AnilistMediaType("ANIME"), search="One Piece")
        result = client._request(client._create_query(vars, *schs), vars)

        # This is equivalent to using the following shortcut:
        media_search = Anilist.QueryClient().media_entry()
        media_search.search(
            Scheme().id,
            Scheme().title.english,

            type=AnilistMediaType("ANIME"),
            search="One Piece"
        )
        result = media_search.results_take_one()
    """

    def __init__(self, client: Client, head_name1, head_name2, defa):
        self._results = []
        self._client = client
        self._head_name = (head_name1, head_name2)
        self._base_schs = (Scheme()[head_name2], Scheme()[head_name1])

        self.DEFAULT_QUERY = defa

    def _query(self, *schs: Scheme, vars, head_sch, paginate=True, page=1, per_page=100):
        r = []

        if not paginate:

            q = self._client._create_query(vars, *schs, head_sch=head_sch)
            resp = self._client._request(q, vars)

            r.append(resp[self._head_name[1]])

        else:

            vars = Vars._merge(vars, Vars(page=page, perPage=per_page))
            q = self._client._create_query(vars, *schs, head_sch=Scheme._combine(Scheme().Page(page = '$page', perPage = '$perPage'), head_sch))

            r = self._client._page_request(q, vars, self._head_name[0])
        
        self._results += r

    
    def get_base_sch(self, paginate):
        return self._base_schs[0] if not paginate else self._base_schs[1]
    
    def search(self, *schs: Scheme, default=True, per_page: int=100, page: int=1, paginate=False, **kwargs):
        """Send a query to the Anilist API and add the results to the result tray

        :param schs: The Schemes to search for,
            they will be appended to the specific Query object's head :class:`~.scheme.Scheme`
        :type schs: tuple[:class:`~.scheme.Scheme`]

        *Keyword Args*: 

        :param default: Whether to add the specific Query object's default :class:`~.scheme.Scheme`\s (the default is True)
        :type default: bool, optional

        :param paginate: Whether to request a single element or pages of the element,
            if set to True then several results will be added to the result tray (the default is False)
        :type paginate: bool, optional

        :param per_page: The number of objects to query per page, this is useless if paginate is set to False (the default is 100)
        :type per_page: int, optional

        :param page: The page to start on (the default is 1)
        :type page: int, optional

        *\*\*kwargs*:
        
        The kwargs pairs are all used to create the search query,
        they are added as arguments to the specific Query object's head :class:`~.scheme.Scheme`

        For example if you are using a :class:`~.query.MediaEntryQuery`:

        .. code-block:: python

            # ... get a MediaEntryQuery
            media_search = ... # media_search is a MediaEntryQuery

            media_search.search(
                # Search schemes
                Scheme().id
                Scheme().status
                Scheme().episodes
                
                # Search parameters
                default=True
                paginate=True
                per_page=50
                page=1

                # Search arguments
                type=AnilistMediaType("ANIME")
                search="One Piece"
            )

            # This generates the following query and variables:
            #
            # Query:
            #
            # query (
            #   $type: MediaType, 
            #   $search: String, 
            #   $perPage: Int, 
            #   $page: Int
            # ) {
            #   Page (page: $page, perPage: $perPage) {
            #       media (type: $type, search: $search) {
            #           id
            #           status
            #           episodes
            #       }
            #    }
            # }
            #
            # Variables:
            #
            # {
            #   "search": "One Piece", 
            #   "type": "ANIME", 
            #   "perPage" 50, 
            #   "page": 1
            # }

        """
        # the main way to interact with this object
        vars = Vars(**kwargs)

        if default:
            schs = list(schs)
            schs.extend(self.DEFAULT_QUERY)

        head_sch = self.get_base_sch(paginate)(**vars._as_query())
        return self._query(*schs, head_sch=head_sch, vars=vars, paginate=paginate, per_page=per_page, page=page)

    def results_take_all(self):
        """Returns the whole result tray and empties it
        """
        r = self._results
        self._results = []
        return r
    
    def results_take_front(self):
        """Returns the first element of the result tray
        and removes it from the tray.
                
        :raises KeyError: If the Query Result is empty

        :return: First element of the query's results tray
        :rtype: AnilistObject
        """
        r = self._results[0]
        self._results = self._results[1:]
        return r
    
    def results_take_back(self):
        """Returns the last element of the Query Result
        and removes it from the Query Result

        :raises KeyError: If the Query Result is empty

        :return: Last element of the query's results tray
        :rtype: :class:`AnilistObject`
        """
        r = self._results[-1]
        self._results = self._results[:-1]
        return r




from Anilist import QueryClient, Scheme, MutationClient, MediaScheme
from Anilist.anilist_types import AnilistMediaType
from Anilist import Auth

from Anilist.scheme import mediaScheme
from Anilist.vars import Vars

import logging
from unittest import TestCase

class ExternalsTest(TestCase):

    def test_auth(self):
        auth = Auth.from_config_file("config.json")
        a = MutationClient(auth, max_pages=10)

        a.media_entry(103572).set_score(35.0)

    
    def test_query(self):
        q = QueryClient(max_pages=10)

        list_query = q.media_list()
        list_query.search(
            # Fields to Query
            mediaScheme().coverImage.medium,
            
            # Search settings
            per_page = 100,
            paginate = True, 

            # Query parameters
            userName="xxxAnn"

        )

        entry_query = q.media_entry()
        entry_query.search(
            # Fields to Query
            Scheme().tags.id,
            Scheme().tags.name,

            # Search settings

            #Query parameters
            type = AnilistMediaType('ANIME'),
            search = "One Piece"

        )
        
        print(list_query.results_take_front())
        print('\n')
        print(entry_query.results_take_front())


import json

from Anilist.auth.token import get_token

class Auth:
    """
    Representation of an authenticated connection to the Anilist API

    :param client_id: The ID of the API Client
    :type client_id: :class:`~int`
        
    :param client_secret: The secret code of the API Client
    :type client_secret: :class:`~str`
    """
    def __init__(self, client_id, client_secret):
        """Constructor method
        """
        self.__id = client_id
        self.__secret = client_secret
        self.__token = None

        self.__gen_token()

    def __gen_token(self):
        self.__token = get_token(self.__id, self.__secret)

    @property
    def token(self):
        return self.__token

    @classmethod
    def from_config_object(cls, config):
        """Authenticates an Auth object from a config object

        :param config: The config object
        :type config: :class:`~dict`

        :return: An authenticated Auth object
        :rtype: :class:`~.Auth`
        """
        return cls(config["ID"], config["SECRET"])
    
    @classmethod
    def from_config_file(cls, file):
        """Authenticates an Auth object from a config file

        :param file: The config file's name
        :type file: :class:`~str`

        :return: An authenticated Auth object
        :rtype: :class:`~.Auth`
        """
        with open(file, 'r') as f:
            config = json.loads(f.read())
        return cls(config["ID"], config["SECRET"])



import requests
import Anilist.utils.consts as consts
from Anilist.auth.code import get_code

def get_token(client_id, client_secret):
    code = get_code(client_id)

    c = requests.post(consts.TOKEN_URI, json={
        'grant_type': 'authorization_code',
        'client_id': client_id,
        'client_secret': client_secret,
        'redirect_uri': 'http://127.0.0.1:8000/',
        'code': code
    }, headers={'Accept': 'application/json'}).json()

    return c["access_token"]    


# define wrappers for Anilist types

class AnilistType:
    def __new__(cls, v):
        if cls == AnilistType:
            if type(v) is str: 
                return AnilistString(None)
            if type(v) is int:
                return AnilistInt(None)
            if type(v) is float:
                return AnilistFloat(None)
            if isinstance(v, AnilistType):
                return v
        else:
            return super().__new__(cls)

    def __str__(self):
        return self.__class__.ANILIST_TYPE_NAME
    
    def __repr__(self):
        return str(self)
    
    def get_inner(self):
        return None

class AnilistString(AnilistType):
    
    ANILIST_TYPE_NAME = "String"

class AnilistInt(AnilistType):

    ANILIST_TYPE_NAME = "Int"

class AnilistFloat(AnilistType):

    ANILIST_TYPE_NAME = "Float"

class AnilistMediaType(AnilistType):
    
    ANILIST_TYPE_NAME = "MediaType"

    def __init__(self, s):
        self._inner = s

    def get_inner(self):
        return self._inner
    




from Anilist.anilist_types import AnilistType

class Vars:
    """
    Set of request variables that can be used in Schemes
    Each variable can be referenced by Schemes by prefixing their name with '$'

    Example::
        
        client = QueryClient()
        # $type and $search
        # reference the Vars' 'type' and 'search' kwargs
        schs = Scheme().Media(type="$type", search="$search").sub_schemes(
            Scheme().id,
            Scheme().title.english
        )
        vars = Vars(type=AnilistMediaType("ANIME"), search="One Piece")
        result = client._request(client._create_query(vars, *schs), vars)

        ## This is equivalent to using the following shortcut:
        media_search = QueryClient().media_entry()
        media_search.search(
            Scheme().id,
            Scheme().title.english,

            type=AnilistMediaType("ANIME"),
            search="One Piece"
        )
        result = media_search.results_take_one()
    """
    
    def __init__(self, **kwargs):
        dic = dict(kwargs)
        l = []
        for k, v in dic.items(): 
            if isinstance(v, AnilistType):
                dic[k] = v.get_inner()
                l.append(f"${k}: {v.__class__.ANILIST_TYPE_NAME}")
            else:
                l.append(f"${k}: {AnilistType(v).__class__.ANILIST_TYPE_NAME}")


        self._dic = dic
        self._vars = l

    def update(self, k, v):
        if isinstance(v, AnilistType):
            self._dic[k] = v.get_inner()
        else:
            self._dic[k] = v

    def _as_query(self):
        temp = {}

        for k, v in self._dic.items():
            temp[k] = f"${k}"

        return temp

    @property
    def _names(self):
        return ', '.join(self._vars)
    
    @property
    def _json(self):
        return self._dic
    
    @classmethod 
    def _merge(cls, vara, varb):
        n = Vars()

        n._vars = vara._vars + varb._vars
        n._dic = vara._dic | varb._dic

        return n



from Anilist.scheme import *
from Anilist.vars import Vars
from Anilist.client.client import Client
from Anilist.query.query import Query
class MediaEntryQuery(Query):
    """
    An interface for making Queries for media to the Anilist API

    
    .. warning::

        This usually should not be created manually, instead use :meth:`QueryClient.media_entry() <.client.QueryClient.media_entry>`

    """

    def __init__(self, client: Client, languages=["english"], sizes=["extraLarge"]):
        super().__init__(client, 'media', 'Media', [
            Scheme().id, 
            *[Scheme().title[lang] for lang in languages], 
            *[Scheme().coverImage[size] for size in sizes]
        ])


from Anilist.scheme import *
from Anilist.vars import Vars
from Anilist.client.client import Client
from Anilist.query.query import Query
class MediaListQuery(Query):
    """
    An interface for making Queries for media lists to the Anilist API

        
    .. warning::

        This usually should not be created manually, instead use :meth:`QueryClient.media_list() <.client.QueryClient.media_list>`

    """
    
    def __init__(self, client: Client, languages=["english"], sizes=["extraLarge"]):
        super().__init__(client, 'mediaList', 'MediaList', [
            mediaScheme().id, 
            *[mediaScheme().title[lang] for lang in languages], 
            *[mediaScheme().coverImage[size] for size in sizes]
        ])
    
    



    
from Anilist.client import Client
from Anilist.query import MediaEntryQuery
from Anilist.query import MediaListQuery

import logging

from Anilist.vars.vars import Vars

class QueryClient(Client):
    """
    A hub that creates various interfaces that interact with the Anilist API

    :param level: The logging level for this client
    :type level: :class:`~int`   
    """

    def __init__(self, level=logging.INFO, **kwargs):
        """Constructor method
        """
        Client.__init__(self, None, level, **kwargs)

    def _query_type(self):
        return "query"

    def media_list(self, *, languages=["english"], sizes=["extraLarge"]):
        """Creates a :class:`~.query.MediaListQuery` object
        """
        return MediaListQuery(client=self, languages=languages, sizes=sizes)
    
    def media_entry(self, *, languages=["english"], sizes=["extraLarge"]):
        """Creates a :class:`~.query.MediaEntryQuery` object
        """
        return MediaEntryQuery(client=self, languages=languages, sizes=sizes)



from Anilist.auth.auth import Auth
from Anilist.client.client import Client
from Anilist.mutation.media_list import MediaEntryMutable

import logging

class MutationClient(Client):

    def __init__(self, auth: Auth, level=logging.INFO, **kwargs):
        Client.__init__(self, auth, level, **kwargs)
    
    def _query_type(self):
        return "mutation"

    def media_entry(self, media_id):
        return MediaEntryMutable._from_media_id(self, media_id)



from Anilist.scheme import Scheme
from unittest import TestCase

class InternalsTest(TestCase):

    def test_scheme(self):
        _ = Scheme._construct(Scheme().A.B(k='$3'), Scheme().A.B.C, Scheme().A.B.C(k="$4").D)


"""
.. highlight:: python

Authentication module

There are several ways to generate an :class:`~.Auth` object::

    # First method
    # Requires a JSON file 
    # with the following structure
    # {
    #   "ID": ...,
    #   "SECRET: ...
    # }
    # where ID and SECRET are the Client ID and the Client Secret

    auth = Auth.from_config_file("filename.json")

    # Second method
    # Requires a dict object
    # with keys "ID" and "SECRET"
    # where ID and SECRET are the Client ID and the Client Secret

    auth = Auth.from_config_object(config_object_name
    
    # Third method
    # where ID and SECRET are the Client ID and the Client Secret

    auth = Auth(ID, SECRET)
"""
from .auth import Auth